"""HUGR builder exceptions."""

from dataclasses import dataclass


@dataclass
class NoSiblingAncestor(Exception):
    """No sibling ancestor of target for valid inter-graph edge."""

    src: int
    tgt: int

    @property
    def msg(self):
        return (
            f"Source {self.src} has no sibling ancestor of target {self.tgt},"
            " so cannot wire up."
        )


@dataclass
class NotInSameCfg(Exception):
    """Source and target nodes are not in the same CFG."""

    src: int
    tgt: int

    @property
    def msg(self):
        return (
            f"Source {self.src} is not in the same CFG as target {self.tgt},"
            " so cannot wire up."
        )


@dataclass
class MismatchedExit(Exception):
    """Edge to exit block signature mismatch."""

    src: int

    @property
    def msg(self):
        return (
            f"Exit branch from node {self.src} does not match existing exit block type."
        )


class ParentBeforeChild(Exception):
    """Parent added before child."""

    msg: str = "Parent node must be added before child node."



"""HUGR edge kinds, types, type parameters and type arguments."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Protocol, runtime_checkable

import hugr.serialization.tys as stys
from hugr.utils import ser_it

ExtensionId = stys.ExtensionId
ExtensionSet = stys.ExtensionSet
TypeBound = stys.TypeBound


class TypeParam(Protocol):
    """A HUGR type parameter."""

    def to_serial(self) -> stys.BaseTypeParam:
        """Convert to serializable model."""
        ...  # pragma: no cover

    def to_serial_root(self) -> stys.TypeParam:
        return stys.TypeParam(root=self.to_serial())  # type: ignore[arg-type]


class TypeArg(Protocol):
    """A HUGR type argument, which can be bound to a :class:TypeParam."""

    def to_serial(self) -> stys.BaseTypeArg:
        """Convert to serializable model."""
        ...  # pragma: no cover

    def to_serial_root(self) -> stys.TypeArg:
        return stys.TypeArg(root=self.to_serial())  # type: ignore[arg-type]


@runtime_checkable
class Type(Protocol):
    """A HUGR type."""

    def to_serial(self) -> stys.BaseType:
        """Convert to serializable model."""
        ...  # pragma: no cover

    def to_serial_root(self) -> stys.Type:
        return stys.Type(root=self.to_serial())  # type: ignore[arg-type]

    def type_arg(self) -> TypeTypeArg:
        """The :class:`TypeTypeArg` for this type.

        Example:
            >>> Qubit.type_arg()
            TypeTypeArg(ty=Qubit)
        """
        return TypeTypeArg(self)


#: Row of types.
TypeRow = list[Type]

# --------------------------------------------
# --------------- TypeParam ------------------
# --------------------------------------------


@dataclass(frozen=True)
class TypeTypeParam(TypeParam):
    """A type parameter indicating a type with a given boumd."""

    bound: TypeBound

    def to_serial(self) -> stys.TypeTypeParam:
        return stys.TypeTypeParam(b=self.bound)


@dataclass(frozen=True)
class BoundedNatParam(TypeParam):
    """A type parameter indicating a natural number with an optional upper bound."""

    upper_bound: int | None

    def to_serial(self) -> stys.BoundedNatParam:
        return stys.BoundedNatParam(bound=self.upper_bound)


@dataclass(frozen=True)
class OpaqueParam(TypeParam):
    """Opaque type parameter."""

    ty: Opaque

    def to_serial(self) -> stys.OpaqueParam:
        return stys.OpaqueParam(ty=self.ty.to_serial())


@dataclass(frozen=True)
class ListParam(TypeParam):
    """Type parameter which requires a list of type arguments."""

    param: TypeParam

    def to_serial(self) -> stys.ListParam:
        return stys.ListParam(param=self.param.to_serial_root())


@dataclass(frozen=True)
class TupleParam(TypeParam):
    """Type parameter which requires a tuple of type arguments."""

    params: list[TypeParam]

    def to_serial(self) -> stys.TupleParam:
        return stys.TupleParam(params=ser_it(self.params))


@dataclass(frozen=True)
class ExtensionsParam(TypeParam):
    """An extension set parameter."""

    def to_serial(self) -> stys.ExtensionsParam:
        return stys.ExtensionsParam()


# ------------------------------------------
# --------------- TypeArg ------------------
# ------------------------------------------


@dataclass(frozen=True)
class TypeTypeArg(TypeArg):
    """A type argument for a :class:`TypeTypeParam`."""

    ty: Type

    def to_serial(self) -> stys.TypeTypeArg:
        return stys.TypeTypeArg(ty=self.ty.to_serial_root())


@dataclass(frozen=True)
class BoundedNatArg(TypeArg):
    """A type argument for a :class:`BoundedNatParam`."""

    n: int

    def to_serial(self) -> stys.BoundedNatArg:
        return stys.BoundedNatArg(n=self.n)


@dataclass(frozen=True)
class OpaqueArg(TypeArg):
    """An opaque type argument for a :class:`OpaqueParam`."""

    ty: Opaque
    value: Any

    def to_serial(self) -> stys.OpaqueArg:
        return stys.OpaqueArg(typ=self.ty.to_serial(), value=self.value)


@dataclass(frozen=True)
class SequenceArg(TypeArg):
    """Sequence of type arguments, for a :class:`ListParam` or :class:`TupleParam`."""

    elems: list[TypeArg]

    def to_serial(self) -> stys.SequenceArg:
        return stys.SequenceArg(elems=ser_it(self.elems))


@dataclass(frozen=True)
class ExtensionsArg(TypeArg):
    """Type argument for an :class:`ExtensionsParam`."""

    extensions: ExtensionSet

    def to_serial(self) -> stys.ExtensionsArg:
        return stys.ExtensionsArg(es=self.extensions)


@dataclass(frozen=True)
class VariableArg(TypeArg):
    """A type argument variable."""

    idx: int
    param: TypeParam

    def to_serial(self) -> stys.VariableArg:
        return stys.VariableArg(idx=self.idx, cached_decl=self.param.to_serial_root())


# ----------------------------------------------
# --------------- Type -------------------------
# ----------------------------------------------


@dataclass(frozen=True)
class Array(Type):
    """Prelude fixed `size` array of `ty` elements."""

    ty: Type
    size: int

    def to_serial(self) -> stys.Array:
        return stys.Array(ty=self.ty.to_serial_root(), len=self.size)


@dataclass()
class Sum(Type):
    """Algebraic sum-over-product type. Instances of this type correspond to
    tuples (products) over one of the `variant_rows` in the sum type, tagged by
    the index of the row.
    """

    variant_rows: list[TypeRow]

    def to_serial(self) -> stys.GeneralSum:
        return stys.GeneralSum(rows=[ser_it(row) for row in self.variant_rows])

    def as_tuple(self) -> Tuple:
        assert (
            len(self.variant_rows) == 1
        ), "Sum type must have exactly one row to be converted to a Tuple"
        return Tuple(*self.variant_rows[0])

    def __repr__(self) -> str:
        return f"Sum({self.variant_rows})"


@dataclass()
class UnitSum(Sum):
    """Simple :class:`Sum` type with `size` variants of empty rows."""

    size: int

    def __init__(self, size: int):
        self.size = size
        super().__init__(variant_rows=[[]] * size)

    def to_serial(self) -> stys.UnitSum:  # type: ignore[override]
        return stys.UnitSum(size=self.size)

    def __repr__(self) -> str:
        if self == Bool:
            return "Bool"
        elif self == Unit:
            return "Unit"
        return f"UnitSum({self.size})"


@dataclass()
class Tuple(Sum):
    """Product type with `tys` elements. Instances of this type correspond to
    :class:`Sum` with a single variant.
    """

    def __init__(self, *tys: Type):
        self.variant_rows = [list(tys)]

    def __repr__(self) -> str:
        return f"Tuple{tuple(self.variant_rows[0])}"


@dataclass(frozen=True)
class Variable(Type):
    """A type variable with a given bound, identified by index."""

    idx: int
    bound: TypeBound

    def to_serial(self) -> stys.Variable:
        return stys.Variable(i=self.idx, b=self.bound)


@dataclass(frozen=True)
class RowVariable(Type):
    """A type variable standing in for a row of types, identified by index."""

    idx: int
    bound: TypeBound

    def to_serial(self) -> stys.RowVar:
        return stys.RowVar(i=self.idx, b=self.bound)


@dataclass(frozen=True)
class USize(Type):
    """The Prelude unsigned size type."""

    def to_serial(self) -> stys.USize:
        return stys.USize()


@dataclass(frozen=True)
class Alias(Type):
    """Type alias."""

    name: str
    bound: TypeBound

    def to_serial(self) -> stys.Alias:
        return stys.Alias(name=self.name, bound=self.bound)


@dataclass(frozen=True)
class FunctionType(Type):
    """A function type, defined by input types,
    output types and extension requirements.
    """

    input: TypeRow
    output: TypeRow
    extension_reqs: ExtensionSet = field(default_factory=ExtensionSet)

    def to_serial(self) -> stys.FunctionType:
        return stys.FunctionType(
            input=ser_it(self.input),
            output=ser_it(self.output),
            extension_reqs=self.extension_reqs,
        )

    @classmethod
    def empty(cls) -> FunctionType:
        """Generate an empty function type.

        Example:
            >>> FunctionType.empty()
            FunctionType([], [])
        """
        return cls(input=[], output=[])

    @classmethod
    def endo(cls, tys: TypeRow) -> FunctionType:
        """Function type with the same input and output types.

        Example:
            >>> FunctionType.endo([Qubit])
            FunctionType([Qubit], [Qubit])
        """
        return cls(input=tys, output=tys)

    def flip(self) -> FunctionType:
        """Return a new function type with input and output types swapped.

        Example:
            >>> FunctionType([Qubit], [Bool]).flip()
            FunctionType([Bool], [Qubit])
        """
        return FunctionType(input=list(self.output), output=list(self.input))

    def __repr__(self) -> str:
        return f"FunctionType({self.input}, {self.output})"


@dataclass(frozen=True)
class PolyFuncType(Type):
    """Polymorphic function type or type scheme. Defined by a list of type
    parameters that may appear in the :class:`FunctionType` body.
    """

    params: list[TypeParam]
    body: FunctionType

    def to_serial(self) -> stys.PolyFuncType:
        return stys.PolyFuncType(
            params=[p.to_serial_root() for p in self.params], body=self.body.to_serial()
        )


@dataclass
class Opaque(Type):
    """Opaque type, identified by `id` and with optional type arguments and bound."""

    id: str
    bound: TypeBound
    args: list[TypeArg] = field(default_factory=list)
    extension: ExtensionId = ""

    def to_serial(self) -> stys.Opaque:
        return stys.Opaque(
            extension=self.extension,
            id=self.id,
            args=[arg.to_serial_root() for arg in self.args],
            bound=self.bound,
        )


@dataclass
class _QubitDef(Type):
    def to_serial(self) -> stys.Qubit:
        return stys.Qubit()

    def __repr__(self) -> str:
        return "Qubit"


#: Qubit type.
Qubit = _QubitDef()
#: Boolean type (:class:`UnitSum` of size 2).
Bool = UnitSum(size=2)
#: Unit type (:class:`UnitSum` of size 1).
Unit = UnitSum(size=1)


@dataclass(frozen=True)
class ValueKind:
    """Dataflow value edges."""

    #: Type of the value.
    ty: Type

    def __repr__(self) -> str:
        return f"ValueKind({self.ty})"


@dataclass(frozen=True)
class ConstKind:
    """Static constant value edges."""

    #: Type of the constant.
    ty: Type

    def __repr__(self) -> str:
        return f"ConstKind({self.ty})"


@dataclass(frozen=True)
class FunctionKind:
    """Statically defined function edges."""

    #: Type of the function.
    ty: PolyFuncType

    def __repr__(self) -> str:
        return f"FunctionKind({self.ty})"


@dataclass(frozen=True)
class CFKind:
    """Control flow edges."""


@dataclass(frozen=True)
class OrderKind:
    """State order edges."""


#: The kind of a HUGR graph edge.
Kind = ValueKind | ConstKind | FunctionKind | CFKind | OrderKind


def get_first_sum(types: TypeRow) -> tuple[Sum, TypeRow]:
    """Check the first type in a row of types is a :class:`Sum`, returning it
    and the rest.

    Args:
        types: row of types.

    Raises:
        AssertionError: if the first type is not a :class:`Sum`.

    Example:
        >>> get_first_sum([UnitSum(3), Qubit])
        (UnitSum(3), [Qubit])
        >>> get_first_sum([Qubit, UnitSum(3)]) # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        AssertionError: Expected Sum, got Qubit
    """
    (sum_, *other) = types
    assert isinstance(sum_, Sum), f"Expected Sum, got {sum_}"
    return sum_, other



"""`hugr` is a Python package for the Quantinuum HUGR common
representation.
"""

# This is updated by our release-please workflow, triggered by this
# annotation: x-release-please-version
__version__ = "0.3.0"



"""Core data structures for HUGR."""

from __future__ import annotations

from collections.abc import Iterable, Mapping
from dataclasses import dataclass, field, replace
from typing import (
    TYPE_CHECKING,
    Generic,
    Protocol,
    TypeVar,
    cast,
    overload,
)

from hugr.node_port import Direction, InPort, Node, OutPort, ToNode, _SubPort
from hugr.ops import Call, Const, DataflowOp, Module, Op
from hugr.serialization.ops import OpType as SerialOp
from hugr.serialization.serial_hugr import SerialHugr
from hugr.tys import Kind, Type, ValueKind
from hugr.utils import BiMap
from hugr.val import Value

from .exceptions import ParentBeforeChild

if TYPE_CHECKING:
    from hugr.val import Value


@dataclass()
class NodeData:
    """Node weights in HUGR graph. Defined by an operation and parent node."""

    #: The operation of the node.
    op: Op
    #: The parent node, or None for the root node.
    parent: Node | None
    _num_inps: int = field(default=0, repr=False)
    _num_outs: int = field(default=0, repr=False)
    children: list[Node] = field(default_factory=list, repr=False)

    def to_serial(self, node: Node) -> SerialOp:
        o = self.op.to_serial(self.parent if self.parent else node)

        return SerialOp(root=o)  # type: ignore[arg-type]


_SO = _SubPort[OutPort]
_SI = _SubPort[InPort]

P = TypeVar("P", InPort, OutPort)
K = TypeVar("K", InPort, OutPort)
OpVar = TypeVar("OpVar", bound=Op)
OpVar2 = TypeVar("OpVar2", bound=Op)


class ParentBuilder(ToNode, Protocol[OpVar]):
    """Abstract interface implemented by builders of nodes that contain child HUGRs."""

    #: The child HUGR.
    hugr: Hugr[OpVar]
    # Unique parent node.
    parent_node: Node

    def to_node(self) -> Node:
        return self.parent_node

    @property
    def parent_op(self) -> OpVar:
        """The parent node's operation."""
        return cast(OpVar, self.hugr[self.parent_node].op)


@dataclass()
class Hugr(Mapping[Node, NodeData], Generic[OpVar]):
    """The core HUGR datastructure.

    Args:
        root_op: The operation for the root node. Defaults to a Module.

    Examples:
        >>> h = Hugr()
        >>> h.root_op()
        Module()
        >>> h[h.root].op
        Module()
    """

    #: Root node of the HUGR.
    root: Node
    # List of nodes, with None for deleted nodes.
    _nodes: list[NodeData | None]
    # Bidirectional map of links between ports.
    _links: BiMap[_SO, _SI]
    # List of free node indices, populated when nodes are deleted.
    _free_nodes: list[Node]

    def __init__(self, root_op: OpVar | None = None) -> None:
        self._free_nodes = []
        self._links = BiMap()
        self._nodes = []
        self.root = self._add_node(root_op or Module(), None, 0)

    def __getitem__(self, key: ToNode) -> NodeData:
        key = key.to_node()
        try:
            n = self._nodes[key.idx]
        except IndexError:
            n = None
        if n is None:
            raise KeyError(key)
        return n

    def __iter__(self):
        return iter(self._nodes)

    def __len__(self) -> int:
        return self.num_nodes()

    def _get_typed_op(self, node: ToNode, cl: type[OpVar2]) -> OpVar2:
        op = self[node].op
        assert isinstance(op, cl)
        return op

    def children(self, node: ToNode | None = None) -> list[Node]:
        """The child nodes of a given `node`.

        Args:
            node: Parent node. Defaults to the HUGR root.

        Returns:
            List of child nodes.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_node(ops.Const(val.TRUE))
            >>> h.children(h.root)
            [Node(1)]
        """
        node = node or self.root
        return self[node].children

    def _add_node(
        self,
        op: Op,
        parent: ToNode | None = None,
        num_outs: int | None = None,
    ) -> Node:
        parent = parent.to_node() if parent else None
        node_data = NodeData(op, parent)

        if self._free_nodes:
            node = self._free_nodes.pop()
            self._nodes[node.idx] = node_data
        else:
            node = Node(len(self._nodes))
            self._nodes.append(node_data)
        node = replace(node, _num_out_ports=num_outs)
        if parent:
            self[parent].children.append(node)
        return node

    def add_node(
        self,
        op: Op,
        parent: ToNode | None = None,
        num_outs: int | None = None,
    ) -> Node:
        """Add a node to the HUGR.

        Args:
            op: Operation of the node.
            parent: Parent node of added node. Defaults to HUGR root if None.
            num_outs: Number of output ports expected for this node. Defaults to None.

        Returns:
            Handle to the added node.
        """
        parent = parent or self.root
        return self._add_node(op, parent, num_outs)

    def add_const(self, value: Value, parent: ToNode | None = None) -> Node:
        """Add a constant node to the HUGR.

        Args:
            value: Value of the constant.
            parent: Parent node of added node. Defaults to HUGR root if None.

        Returns:
            Handle to the added node.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_const(val.TRUE)
            >>> h[n].op
            Const(TRUE)
        """
        return self.add_node(Const(value), parent)

    def delete_node(self, node: ToNode) -> NodeData | None:
        """Delete a node from the HUGR.

        Args:
            node: Node to delete.

        Returns:
            The deleted node data, or None if the node was not found.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_const(val.TRUE)
            >>> deleted = h.delete_node(n)
            >>> deleted.op
            Const(TRUE)
            >>> len(h)
            1
        """
        node = node.to_node()
        parent = self[node].parent
        if parent:
            self[parent].children.remove(node)
        for offset in range(self.num_in_ports(node)):
            self._links.delete_right(_SubPort(node.inp(offset)))
        for offset in range(self.num_out_ports(node)):
            self._links.delete_left(_SubPort(node.out(offset)))

        weight, self._nodes[node.idx] = self._nodes[node.idx], None
        self._free_nodes.append(node)
        return weight

    def _unused_sub_offset(self, port: P) -> _SubPort[P]:
        d: dict[_SO, _SI] | dict[_SI, _SO]
        match port:
            case OutPort(_):
                d = self._links.fwd
            case InPort(_):
                d = self._links.bck
        sub_port = _SubPort(port)
        while sub_port in d:
            sub_port = sub_port.next_sub_offset()
        return sub_port

    def add_link(self, src: OutPort, dst: InPort) -> None:
        """Add a link (edge) between two nodes to the HUGR,
          from an outgoing port to an incoming port.

        Args:
            src: Source port.
            dst: Destination port.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> list(df.hugr.linked_ports(df.input_node[0]))
            [InPort(Node(2), 0)]
        """
        src_sub = self._unused_sub_offset(src)
        dst_sub = self._unused_sub_offset(dst)
        # if self._links.get_left(dst_sub) is not None:
        #     dst = replace(dst, _sub_offset=dst._sub_offset + 1)
        self._links.insert_left(src_sub, dst_sub)

        self[src.node]._num_outs = max(self[src.node]._num_outs, src.offset + 1)
        self[dst.node]._num_inps = max(self[dst.node]._num_inps, dst.offset + 1)

    def delete_link(self, src: OutPort, dst: InPort) -> None:
        """Delete a link (edge) between two nodes from the HUGR.

        Args:
            src: Source port.
            dst: Destination port.
        """
        try:
            sub_offset = next(
                i for i, inp in enumerate(self.linked_ports(src)) if inp == dst
            )
            self._links.delete_left(_SubPort(src, sub_offset))
        except StopIteration:
            return
        # TODO make sure sub-offset is handled correctly

    def root_op(self) -> OpVar:
        """The operation of the root node.

        Examples:
            >>> h = Hugr()
            >>> h.root_op()
            Module()
        """
        return cast(OpVar, self[self.root].op)

    def num_nodes(self) -> int:
        """The number of nodes in the HUGR.

        Examples:
            >>> h = Hugr()
            >>> n = h.add_const(val.TRUE)
            >>> h.num_nodes()
            2
        """
        return len(self._nodes) - len(self._free_nodes)

    def num_ports(self, node: ToNode, direction: Direction) -> int:
        """The number of ports of a node in a given direction.
        Not necessarily the number of connected ports - if port `i` is
        connected, then all ports `0..i` are assumed to exist.

        Args:
            node: Node to query.
            direction: Direction of ports to count.

        Examples:
            >>> h = Hugr()
            >>> n1 = h.add_const(val.TRUE)
            >>> n2 = h.add_const(val.FALSE)
            >>> h.add_link(n1.out(0), n2.inp(2)) # not a valid link!
            >>> h.num_ports(n1, Direction.OUTGOING)
            1
            >>> h.num_ports(n2, Direction.INCOMING)
            3
        """
        return (
            self.num_in_ports(node)
            if direction == Direction.INCOMING
            else self.num_out_ports(node)
        )

    def num_in_ports(self, node: ToNode) -> int:
        """The number of incoming ports of a node. See :meth:`num_ports`."""
        return self[node]._num_inps

    def num_out_ports(self, node: ToNode) -> int:
        """The number of outgoing ports of a node. See :meth:`num_ports`."""
        return self[node]._num_outs

    def _linked_ports(
        self, port: P, links: dict[_SubPort[P], _SubPort[K]]
    ) -> Iterable[K]:
        sub_port = _SubPort(port)
        while sub_port in links:
            # sub offset not used in API
            yield links[sub_port].port
            sub_port = sub_port.next_sub_offset()

    @overload
    def linked_ports(self, port: OutPort) -> Iterable[InPort]: ...
    @overload
    def linked_ports(self, port: InPort) -> Iterable[OutPort]: ...
    def linked_ports(self, port: OutPort | InPort):
        """Return an iterable of In(Out)Ports linked to given Out(In)Port.

        Args:
            port: Given port.

        Returns:
            Iterator over linked ports.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.set_outputs(df.input_node[0])
            >>> list(df.hugr.linked_ports(df.input_node[0]))
            [InPort(Node(2), 0)]

        """
        match port:
            case OutPort(_):
                return self._linked_ports(port, self._links.fwd)
            case InPort(_):
                return self._linked_ports(port, self._links.bck)

    # TODO: single linked port

    def outgoing_order_links(self, node: ToNode) -> Iterable[Node]:
        """Iterator over nodes connected by an outgoing state order link from a
        given node.

        Args:
            node: Source node of state order link.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.add_state_order(df.input_node, df.output_node)
            >>> list(df.hugr.outgoing_order_links(df.input_node))
            [Node(2)]
        """
        return (p.node for p in self.linked_ports(node.out(-1)))

    def incoming_order_links(self, node: ToNode) -> Iterable[Node]:
        """Iterator over nodes connected by an incoming state order link to a
        given node.

        Args:
            node: Destination node of state order link.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.add_state_order(df.input_node, df.output_node)
            >>> list(df.hugr.incoming_order_links(df.output_node))
            [Node(1)]
        """
        return (p.node for p in self.linked_ports(node.inp(-1)))

    def _node_links(
        self, node: ToNode, links: dict[_SubPort[P], _SubPort[K]]
    ) -> Iterable[tuple[P, list[K]]]:
        try:
            direction = next(iter(links.keys())).port.direction
        except StopIteration:
            return
        # iterate over known offsets
        for offset in range(self.num_ports(node, direction)):
            port = cast(P, node.port(offset, direction))
            yield port, list(self._linked_ports(port, links))

    def outgoing_links(self, node: ToNode) -> Iterable[tuple[OutPort, list[InPort]]]:
        """Iterator over outgoing links from a given node.

        Args:
            node: Node to query.

        Returns:
            Iterator of pairs of outgoing port and the incoming ports connected
            to that port.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> list(df.hugr.outgoing_links(df.input_node))
            [(OutPort(Node(1), 0), [InPort(Node(2), 0), InPort(Node(2), 1)])]
        """
        return self._node_links(node, self._links.fwd)

    def incoming_links(self, node: ToNode) -> Iterable[tuple[InPort, list[OutPort]]]:
        """Iterator over incoming links to a given node.

        Args:
            node: Node to query.

        Returns:
            Iterator of pairs of incoming port and the outgoing ports connected
            to that port.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(1))
            >>> list(df.hugr.incoming_links(df.output_node))
            [(InPort(Node(2), 0), [OutPort(Node(1), 0)]), (InPort(Node(2), 1), [OutPort(Node(1), 0)])]
        """  # noqa: E501
        return self._node_links(node, self._links.bck)

    def num_incoming(self, node: Node) -> int:
        """The number of incoming links to a `node`.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.num_incoming(df.output_node)
            1
        """
        return sum(1 for _ in self.incoming_links(node))

    def num_outgoing(self, node: ToNode) -> int:
        """The number of outgoing links from a `node`.

        Examples:
            >>> df = dfg.Dfg()
            >>> df.hugr.add_link(df.input_node.out(0), df.output_node.inp(0))
            >>> df.hugr.num_outgoing(df.input_node)
            1
        """
        return sum(1 for _ in self.outgoing_links(node))

    # TODO: num_links and _linked_ports

    def port_kind(self, port: InPort | OutPort) -> Kind:
        """The kind of a `port`.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.port_kind(df.input_node.out(0))
            ValueKind(Bool)
        """
        return self[port.node].op.port_kind(port)

    def port_type(self, port: InPort | OutPort) -> Type | None:
        """The type of a `port`, if the kind is
        :class:`ValueKind <hugr.tys.ValueKind>`, else None.

        Examples:
            >>> df = dfg.Dfg(tys.Bool)
            >>> df.hugr.port_type(df.input_node.out(0))
            Bool
        """
        op = self[port.node].op
        if isinstance(op, DataflowOp):
            return op.port_type(port)
        if isinstance(op, Call) and isinstance(port, OutPort):
            kind = self.port_kind(port)
            if isinstance(kind, ValueKind):
                return kind.ty
        return None

    def insert_hugr(self, hugr: Hugr, parent: ToNode | None = None) -> dict[Node, Node]:
        """Insert a HUGR into this HUGR.

        Args:
            hugr: HUGR to insert.
            parent: Parent for root of inserted HUGR. Defaults to None.

        Returns:
            Mapping from node indices in inserted HUGR to their new indices
            in this HUGR.

        Examples:
            >>> d = dfg.Dfg()
            >>> h = Hugr()
            >>> h.insert_hugr(d.hugr)
            {Node(0): Node(1), Node(1): Node(2), Node(2): Node(3)}
        """
        mapping: dict[Node, Node] = {}

        for idx, node_data in enumerate(hugr._nodes):
            if node_data is not None:
                # relies on parents being inserted before any children
                try:
                    node_parent = (
                        mapping[node_data.parent] if node_data.parent else parent
                    )
                except KeyError as e:
                    raise ParentBeforeChild from e
                mapping[Node(idx)] = self.add_node(node_data.op, node_parent)

        for src, dst in hugr._links.items():
            self.add_link(
                mapping[src.port.node].out(src.port.offset),
                mapping[dst.port.node].inp(dst.port.offset),
            )
        return mapping

    def to_serial(self) -> SerialHugr:
        """Serialize the HUGR."""
        node_it = (node for node in self._nodes if node is not None)

        def _serialize_link(
            link: tuple[_SO, _SI],
        ) -> tuple[tuple[int, int], tuple[int, int]]:
            src, dst = link
            s, d = self._constrain_offset(src.port), self._constrain_offset(dst.port)
            return (src.port.node.idx, s), (dst.port.node.idx, d)

        return SerialHugr(
            version="v1",
            # non contiguous indices will be erased
            nodes=[node.to_serial(Node(idx)) for idx, node in enumerate(node_it)],
            edges=[_serialize_link(link) for link in self._links.items()],
        )

    def _constrain_offset(self, p: P) -> int:
        # negative offsets are used to refer to the last port
        if p.offset < 0:
            match p.direction:
                case Direction.INCOMING:
                    current = self.num_incoming(p.node)
                case Direction.OUTGOING:
                    current = self.num_outgoing(p.node)
            offset = current + p.offset + 1
        else:
            offset = p.offset

        return offset

    @classmethod
    def from_serial(cls, serial: SerialHugr) -> Hugr:
        """Load a HUGR from a serialized form."""
        assert serial.nodes, "Empty Hugr is invalid"

        hugr = Hugr.__new__(Hugr)
        hugr._nodes = []
        hugr._links = BiMap()
        hugr._free_nodes = []
        hugr.root = Node(0)
        for idx, serial_node in enumerate(serial.nodes):
            parent: Node | None = Node(serial_node.root.parent)
            if serial_node.root.parent == idx:
                hugr.root = Node(idx)
                parent = None
            serial_node.root.parent = -1
            hugr._nodes.append(NodeData(serial_node.root.deserialize(), parent))

        for (src_node, src_offset), (dst_node, dst_offset) in serial.edges:
            if src_offset is None or dst_offset is None:
                continue
            hugr.add_link(
                Node(src_node).out(src_offset), Node(dst_node).inp(dst_offset)
            )

        return hugr



"""HUGR values, used for static constants in HUGR programs."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Protocol, runtime_checkable

import hugr.serialization.ops as sops
import hugr.serialization.tys as stys
from hugr import tys
from hugr.utils import ser_it

if TYPE_CHECKING:
    from hugr.hugr import Hugr


@runtime_checkable
class Value(Protocol):
    """Abstract value definition. Must be serializable into a HUGR value."""

    def to_serial(self) -> sops.BaseValue:
        """Convert to serializable model."""
        ...  # pragma: no cover

    def to_serial_root(self) -> sops.Value:
        return sops.Value(root=self.to_serial())  # type: ignore[arg-type]

    def type_(self) -> tys.Type:
        """Report the type of the value.

        Example:
            >>> TRUE.type_()
            Bool
        """
        ...  # pragma: no cover


@dataclass
class Sum(Value):
    """Sum-of-product value.

    Example:
        >>> Sum(0, tys.Sum([[tys.Bool], [tys.Unit]]), [TRUE])
        Sum(tag=0, typ=Sum([[Bool], [Unit]]), vals=[TRUE])
    """

    #: Tag identifying the variant.
    tag: int
    #: Type of the sum: defines all possible variants.
    typ: tys.Sum
    #: The values of this variant row.
    vals: list[Value]

    @property
    def n_variants(self) -> int:
        return len(self.typ.variant_rows)

    def type_(self) -> tys.Sum:
        return self.typ

    def to_serial(self) -> sops.SumValue:
        return sops.SumValue(
            tag=self.tag,
            typ=stys.SumType(root=self.type_().to_serial()),
            vs=ser_it(self.vals),
        )


class UnitSum(Sum):
    """Simple :class:`Sum` with each variant being an empty row.

    Example:
        >>> UnitSum(0, 3)
        UnitSum(0, 3)
        >>> UnitSum(0, 1)
        Unit
        >>> assert UnitSum(0, 2) == FALSE
        >>> assert UnitSum(1, 2) == TRUE
    """

    def __init__(self, tag: int, size: int):
        super().__init__(
            tag=tag,
            typ=tys.UnitSum(size),
            vals=[],
        )

    def __repr__(self) -> str:
        if self == TRUE:
            return "TRUE"
        if self == FALSE:
            return "FALSE"
        if self == Unit:
            return "Unit"
        return f"UnitSum({self.tag}, {self.n_variants})"


def bool_value(b: bool) -> UnitSum:
    """Convert a python bool to a HUGR boolean value.

    Example:
        >>> bool_value(True)
        TRUE
        >>> bool_value(False)
        FALSE
    """
    return UnitSum(int(b), 2)


#: HUGR unit type. Sum with a single empty row variant.
Unit = UnitSum(0, 1)
#: HUGR true value.
TRUE = bool_value(True)
#: HUGR false value.
FALSE = bool_value(False)


@dataclass
class Tuple(Sum):
    """Tuple or product value, defined by a list of values.
    Internally a :class:`Sum` with a single variant row.

    Example:
        >>> tup = Tuple(TRUE, FALSE)
        >>> tup
        Tuple(TRUE, FALSE)
        >>> tup.type_()
        Tuple(Bool, Bool)

    """

    #: The values of this tuple.
    vals: list[Value]

    def __init__(self, *vals: Value):
        val_list = list(vals)
        super().__init__(
            tag=0, typ=tys.Tuple(*(v.type_() for v in val_list)), vals=val_list
        )

    # sops.TupleValue isn't an instance of sops.SumValue
    # so mypy doesn't like the override of Sum.to_serial
    def to_serial(self) -> sops.TupleValue:  # type: ignore[override]
        return sops.TupleValue(
            vs=ser_it(self.vals),
        )

    def __repr__(self) -> str:
        return f"Tuple({', '.join(map(repr, self.vals))})"


@dataclass
class Function(Value):
    """Higher order function value, defined by a :class:`Hugr <hugr.hugr.HUGR>`."""

    body: Hugr

    def type_(self) -> tys.FunctionType:
        return self.body.root_op().inner_signature()

    def to_serial(self) -> sops.FunctionValue:
        return sops.FunctionValue(
            hugr=self.body.to_serial(),
        )


@dataclass
class Extension(Value):
    """Non-core extension value."""

    #: Value name.
    name: str
    #: Value type.
    typ: tys.Type
    #: Value payload.
    val: Any
    extensions: tys.ExtensionSet = field(default_factory=tys.ExtensionSet)

    def type_(self) -> tys.Type:
        return self.typ

    def to_serial(self) -> sops.ExtensionValue:
        return sops.ExtensionValue(
            typ=self.typ.to_serial_root(),
            value=sops.CustomConst(c=self.name, v=self.val),
            extensions=self.extensions,
        )


class ExtensionValue(Value, Protocol):
    """Protocol which types can implement to be a HUGR extension value."""

    def to_value(self) -> Extension:
        """Convert to a HUGR extension value."""
        ...  # pragma: no cover

    def type_(self) -> tys.Type:
        return self.to_value().type_()

    def to_serial(self) -> sops.ExtensionValue:
        return self.to_value().to_serial()



"""Definitions of HUGR operations."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Protocol, TypeVar, runtime_checkable

import hugr.serialization.ops as sops
from hugr import tys, val
from hugr.node_port import Direction, InPort, Node, OutPort, Wire
from hugr.utils import ser_it

if TYPE_CHECKING:
    from collections.abc import Sequence

    from hugr.serialization.ops import BaseOp


@dataclass
class InvalidPort(Exception):
    """Port is not valid for this operation."""

    port: InPort | OutPort
    op: Op

    @property
    def msg(self) -> str:
        return f"Port {self.port} is invalid for operation {self.op}."


@runtime_checkable
class Op(Protocol):
    """An abstract HUGR operation. Must be convertible
    to a serializable :class:`BaseOp`.
    """

    @property
    def num_out(self) -> int:
        """The number of output ports for this operation.

        Example:
            >>> op = Const(val.TRUE)
            >>> op.num_out
            1
        """
        ...  # pragma: no cover

    def to_serial(self, parent: Node) -> BaseOp:
        """Convert this operation to a serializable form."""
        ...  # pragma: no cover

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        """Get the kind of the given port.

        Example:
            >>> op = Const(val.TRUE)
            >>> op.port_kind(OutPort(Node(0), 0))
            ConstKind(Bool)
        """
        ...  # pragma: no cover

    def _invalid_port(self, port: InPort | OutPort) -> InvalidPort:
        return InvalidPort(port, self)


def _sig_port_type(sig: tys.FunctionType, port: InPort | OutPort) -> tys.Type:
    if port.direction == Direction.INCOMING:
        return sig.input[port.offset]
    return sig.output[port.offset]


@runtime_checkable
class DataflowOp(Op, Protocol):
    """Abstract dataflow operation. Can be assumed to have a signature and Value-
    kind ports.
    """

    def outer_signature(self) -> tys.FunctionType:
        """The external signature of this operation. Defines the valid external
        connectivity of the node the operation belongs to.
        """
        ...  # pragma: no cover

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        if port.offset == -1:
            return tys.OrderKind()
        return tys.ValueKind(self.port_type(port))

    def port_type(self, port: InPort | OutPort) -> tys.Type:
        """Get the type of the given dataflow port from the signature of the
        operation.

        Example:
            >>> op = Input([tys.Bool])
            >>> op.port_type(OutPort(Node(0), 0))
            Bool

        """
        return _sig_port_type(self.outer_signature(), port)

    def __call__(self, *args) -> Command:
        """Calling with incoming :class:`Wire` arguments returns a
        :class:`Command` which can be used to wire the operation into a
        dataflow graph.
        """
        return Command(self, list(args))


@runtime_checkable
class _PartialOp(Protocol):
    def _set_in_types(self, types: tys.TypeRow) -> None: ...


@dataclass
class IncompleteOp(Exception):
    """Op types have not been set during building."""

    op: Op

    @property
    def msg(self) -> str:
        return (
            f"Operation {self.op} is incomplete, may require set_in_types to be called."
        )


V = TypeVar("V")


def _check_complete(op, v: V | None) -> V:
    if v is None:
        raise IncompleteOp(op)
    return v


@dataclass(frozen=True)
class Command:
    """A :class:`DataflowOp` and its incoming :class:`Wire <hugr.nodeport.Wire>`
      arguments.

    Ephemeral: used to wire operations into a dataflow graph.

    Example:
        >>> Noop()(Node(0).out(0))
        Command(op=Noop, incoming=[OutPort(Node(0), 0)])
    """

    op: DataflowOp
    incoming: list[Wire]


@dataclass()
class Input(DataflowOp):
    """Input operation in dataflow graph. Outputs of this operation are the
    inputs to the graph.
    """

    types: tys.TypeRow

    @property
    def num_out(self) -> int:
        return len(self.types)

    def to_serial(self, parent: Node) -> sops.Input:
        return sops.Input(parent=parent.idx, types=ser_it(self.types))

    def outer_signature(self) -> tys.FunctionType:
        return tys.FunctionType(input=[], output=self.types)

    def __call__(self) -> Command:
        return super().__call__()


@dataclass()
class Output(DataflowOp, _PartialOp):
    """Output operation in dataflow graph. Inputs of this operation are the
    outputs of the graph.
    """

    _types: tys.TypeRow | None = field(default=None, repr=False)
    num_out: int = field(default=0, repr=False)

    @property
    def types(self) -> tys.TypeRow:
        return _check_complete(self, self._types)

    def to_serial(self, parent: Node) -> sops.Output:
        return sops.Output(parent=parent.idx, types=ser_it(self.types))

    def outer_signature(self) -> tys.FunctionType:
        return tys.FunctionType(input=self.types, output=[])

    def _set_in_types(self, types: tys.TypeRow) -> None:
        self._types = types


@dataclass()
class Custom(DataflowOp):
    """A non-core dataflow operation defined in an extension."""

    op_name: str
    signature: tys.FunctionType = field(default_factory=tys.FunctionType.empty)
    description: str = ""
    extension: tys.ExtensionId = ""
    args: list[tys.TypeArg] = field(default_factory=list)

    @property
    def num_out(self) -> int:
        return len(self.signature.output)

    def to_serial(self, parent: Node) -> sops.CustomOp:
        return sops.CustomOp(
            parent=parent.idx,
            extension=self.extension,
            op_name=self.op_name,
            signature=self.signature.to_serial(),
            description=self.description,
            args=ser_it(self.args),
        )

    def outer_signature(self) -> tys.FunctionType:
        return self.signature


@dataclass()
class MakeTuple(DataflowOp, _PartialOp):
    """Operation to create a tuple from a sequence of wires."""

    _types: tys.TypeRow | None = field(default=None, repr=False)
    num_out: int = field(default=1, repr=False)

    @property
    def types(self) -> tys.TypeRow:
        """If set, the types of the tuple elements.

        Raises:
            IncompleteOp: If the types have not been set.
        """
        return _check_complete(self, self._types)

    def to_serial(self, parent: Node) -> sops.MakeTuple:
        return sops.MakeTuple(
            parent=parent.idx,
            tys=ser_it(self.types),
        )

    def __call__(self, *elements: Wire) -> Command:
        return super().__call__(*elements)

    def outer_signature(self) -> tys.FunctionType:
        return tys.FunctionType(input=self.types, output=[tys.Tuple(*self.types)])

    def _set_in_types(self, types: tys.TypeRow) -> None:
        self._types = types

    def __repr__(self) -> str:
        return "MakeTuple" + (f"({self._types})" if self._types is not None else "")


@dataclass()
class UnpackTuple(DataflowOp, _PartialOp):
    """Operation to unpack a tuple into its elements."""

    _types: tys.TypeRow | None = field(default=None, repr=False)

    @property
    def types(self) -> tys.TypeRow:
        """If set, the types of the tuple elements.

        Raises:
            IncompleteOp: If the types have not been set.
        """
        return _check_complete(self, self._types)

    @property
    def num_out(self) -> int:
        return len(self.types)

    def to_serial(self, parent: Node) -> sops.UnpackTuple:
        return sops.UnpackTuple(
            parent=parent.idx,
            tys=ser_it(self.types),
        )

    def __call__(self, tuple_: Wire) -> Command:
        return super().__call__(tuple_)

    def outer_signature(self) -> tys.FunctionType:
        return MakeTuple(self.types).outer_signature().flip()

    def _set_in_types(self, types: tys.TypeRow) -> None:
        (t,) = types
        assert isinstance(t, tys.Sum), f"Expected unary Sum, got {t}"
        (row,) = t.variant_rows
        self._types = row


@dataclass()
class Tag(DataflowOp):
    """Tag a row of incoming values to make them a variant of a sum type.

    Requires `sum_ty` to be set as it is not possible to extract all the variants from
    just the input wires for one variant.
    """

    tag: int
    sum_ty: tys.Sum
    num_out: int = field(default=1, repr=False)

    def to_serial(self, parent: Node) -> sops.Tag:
        return sops.Tag(
            parent=parent.idx,
            tag=self.tag,
            variants=[ser_it(r) for r in self.sum_ty.variant_rows],
        )

    def outer_signature(self) -> tys.FunctionType:
        return tys.FunctionType(
            input=self.sum_ty.variant_rows[self.tag], output=[self.sum_ty]
        )


class DfParentOp(Op, Protocol):
    """Abstract parent of dataflow graph operations. Can be queried for the
    dataflow signature of its child graph.
    """

    def inner_signature(self) -> tys.FunctionType:
        """Inner signature of the child dataflow graph."""
        ...  # pragma: no cover

    def _set_out_types(self, types: tys.TypeRow) -> None: ...

    def _inputs(self) -> tys.TypeRow: ...


@dataclass
class DFG(DfParentOp, DataflowOp):
    """Simple dataflow graph operation. Outer signature matches inner signature."""

    #: Inputs types of the operation.
    inputs: tys.TypeRow
    _outputs: tys.TypeRow | None = field(default=None, repr=False)
    _extension_delta: tys.ExtensionSet = field(default_factory=list, repr=False)

    @property
    def outputs(self) -> tys.TypeRow:
        """Output types of the operation.

        Raises:
            IncompleteOp: If the outputs have not been set.
        """
        return _check_complete(self, self._outputs)

    @property
    def signature(self) -> tys.FunctionType:
        """Signature of the operation.

        Raises:
            IncompleteOp: If the outputs have not been set.
        """
        return tys.FunctionType(self.inputs, self.outputs, self._extension_delta)

    @property
    def num_out(self) -> int:
        return len(self.signature.output)

    def to_serial(self, parent: Node) -> sops.DFG:
        return sops.DFG(
            parent=parent.idx,
            signature=self.signature.to_serial(),
        )

    def inner_signature(self) -> tys.FunctionType:
        return self.signature

    def outer_signature(self) -> tys.FunctionType:
        return self.signature

    def _set_out_types(self, types: tys.TypeRow) -> None:
        self._outputs = types

    def _inputs(self) -> tys.TypeRow:
        return self.inputs


@dataclass()
class CFG(DataflowOp):
    """Parent operation of a control flow graph."""

    #: Inputs types of the operation.
    inputs: tys.TypeRow
    _outputs: tys.TypeRow | None = field(default=None, repr=False)

    @property
    def outputs(self) -> tys.TypeRow:
        """Output types of the operation, if set.

        Raises:
            IncompleteOp: If the outputs have not been set.
        """
        return _check_complete(self, self._outputs)

    @property
    def signature(self) -> tys.FunctionType:
        """Dataflow signature of the CFG operation.

        Raises:
            IncompleteOp: If the outputs have not been set.
        """
        return tys.FunctionType(self.inputs, self.outputs)

    @property
    def num_out(self) -> int:
        return len(self.outputs)

    def to_serial(self, parent: Node) -> sops.CFG:
        return sops.CFG(
            parent=parent.idx,
            signature=self.signature.to_serial(),
        )

    def outer_signature(self) -> tys.FunctionType:
        return self.signature


@dataclass
class DataflowBlock(DfParentOp):
    """Parent of non-entry basic block in a control flow graph."""

    #: Inputs types of the innner dataflow graph.
    inputs: tys.TypeRow
    _sum: tys.Sum | None = None
    _other_outputs: tys.TypeRow | None = field(default=None, repr=False)
    extension_delta: tys.ExtensionSet = field(default_factory=list)

    @property
    def sum_ty(self) -> tys.Sum:
        """If set, the sum type that defines the potential branching of the
        block.


        Raises:
            IncompleteOp: If the sum type has not been set.
        """
        return _check_complete(self, self._sum)

    @property
    def other_outputs(self) -> tys.TypeRow:
        """The non-branching outputs of the block which are passed to all
        successors.

        Raises:
            IncompleteOp: If the outputs have not been set.
        """
        return _check_complete(self, self._other_outputs)

    @property
    def num_out(self) -> int:
        return len(self.sum_ty.variant_rows)

    def to_serial(self, parent: Node) -> sops.DataflowBlock:
        return sops.DataflowBlock(
            parent=parent.idx,
            inputs=ser_it(self.inputs),
            sum_rows=list(map(ser_it, self.sum_ty.variant_rows)),
            other_outputs=ser_it(self.other_outputs),
            extension_delta=self.extension_delta,
        )

    def inner_signature(self) -> tys.FunctionType:
        return tys.FunctionType(
            input=self.inputs, output=[self.sum_ty, *self.other_outputs]
        )

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        return tys.CFKind()

    def _set_out_types(self, types: tys.TypeRow) -> None:
        (sum_, other) = tys.get_first_sum(types)
        self._sum = sum_
        self._other_outputs = other

    def _inputs(self) -> tys.TypeRow:
        return self.inputs

    def nth_outputs(self, n: int) -> tys.TypeRow:
        """The outputs passed to the nth successor of the block.
        Concatenation of the nth variant of the sum type and the other outputs.
        """
        return [*self.sum_ty.variant_rows[n], *self.other_outputs]


@dataclass
class ExitBlock(Op):
    """Unique exit block of a control flow graph."""

    _cfg_outputs: tys.TypeRow | None = field(default=None, repr=False)
    num_out: int = field(default=0, repr=False)

    @property
    def cfg_outputs(self) -> tys.TypeRow:
        """Output types of the parent control flow graph of this exit block.

        Raises:
            IncompleteOp: If the outputs have not been set.
        """
        return _check_complete(self, self._cfg_outputs)

    def to_serial(self, parent: Node) -> sops.ExitBlock:
        return sops.ExitBlock(
            parent=parent.idx,
            cfg_outputs=ser_it(self.cfg_outputs),
        )

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        return tys.CFKind()


@dataclass
class Const(Op):
    """A static constant value. Can be used with a :class:`LoadConst` to load into
    a dataflow graph.
    """

    val: val.Value
    num_out: int = field(default=1, repr=False)

    def to_serial(self, parent: Node) -> sops.Const:
        return sops.Const(
            parent=parent.idx,
            v=self.val.to_serial_root(),
        )

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        match port:
            case OutPort(_, 0):
                return tys.ConstKind(self.val.type_())
            case _:
                raise self._invalid_port(port)

    def __repr__(self) -> str:
        return f"Const({self.val})"


@dataclass
class LoadConst(DataflowOp):
    """Load a constant value into a dataflow graph. Connects to a :class:`Const`."""

    _typ: tys.Type | None = None
    num_out: int = field(default=1, repr=False)

    @property
    def type_(self) -> tys.Type:
        """The type of the loaded value.

        Raises:
            IncompleteOp: If the type has not been set.
        """
        return _check_complete(self, self._typ)

    def to_serial(self, parent: Node) -> sops.LoadConstant:
        return sops.LoadConstant(
            parent=parent.idx,
            datatype=self.type_.to_serial_root(),
        )

    def outer_signature(self) -> tys.FunctionType:
        return tys.FunctionType(input=[], output=[self.type_])

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        match port:
            case InPort(_, 0):
                return tys.ConstKind(self.type_)
            case OutPort(_, 0):
                return tys.ValueKind(self.type_)
            case _:
                raise self._invalid_port(port)

    def __repr__(self) -> str:
        return "LoadConst" + (f"({self._typ})" if self._typ is not None else "")


@dataclass()
class Conditional(DataflowOp):
    """'Switch' operation on the variants of an incoming sum type, evaluating the
    corresponding one of the child :class:`Case` operations.
    """

    #: Sum type to switch on.
    sum_ty: tys.Sum
    #: Non-sum inputs that are passed to all cases.
    other_inputs: tys.TypeRow
    _outputs: tys.TypeRow | None = field(default=None, repr=False)

    @property
    def outputs(self) -> tys.TypeRow:
        """Outputs of the conditional, common to all cases.

        Raises:
            IncompleteOp: If the outputs have not been set.
        """
        return _check_complete(self, self._outputs)

    @property
    def signature(self) -> tys.FunctionType:
        """Dataflow signature of the conditional operation.

        Raises:
            IncompleteOp: If the outputs have not been set.
        """
        inputs = [self.sum_ty, *self.other_inputs]
        return tys.FunctionType(inputs, self.outputs)

    @property
    def num_out(self) -> int:
        return len(self.outputs)

    def to_serial(self, parent: Node) -> sops.Conditional:
        return sops.Conditional(
            parent=parent.idx,
            sum_rows=[ser_it(r) for r in self.sum_ty.variant_rows],
            other_inputs=ser_it(self.other_inputs),
            outputs=ser_it(self.outputs),
        )

    def outer_signature(self) -> tys.FunctionType:
        return self.signature

    def nth_inputs(self, n: int) -> tys.TypeRow:
        """The inputs passed to the nth child case.
        Concatenation of the nth variant of the sum type and the other inputs.
        """
        return [*self.sum_ty.variant_rows[n], *self.other_inputs]


@dataclass
class Case(DfParentOp):
    """Parent of a dataflow graph that is a branch of a :class:`Conditional`."""

    #: Inputs types of the innner dataflow graph.
    inputs: tys.TypeRow
    _outputs: tys.TypeRow | None = field(default=None, repr=False)
    num_out: int = field(default=0, repr=False)

    @property
    def outputs(self) -> tys.TypeRow:
        """Outputs of the case operation.

        Raises:
            IncompleteOp: If the outputs have not been set.
        """
        return _check_complete(self, self._outputs)

    def to_serial(self, parent: Node) -> sops.Case:
        return sops.Case(
            parent=parent.idx, signature=self.inner_signature().to_serial()
        )

    def inner_signature(self) -> tys.FunctionType:
        return tys.FunctionType(self.inputs, self.outputs)

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        raise self._invalid_port(port)

    def _set_out_types(self, types: tys.TypeRow) -> None:
        self._outputs = types

    def _inputs(self) -> tys.TypeRow:
        return self.inputs


@dataclass
class TailLoop(DfParentOp, DataflowOp):
    """Tail controlled loop operation, child dataflow graph iterates while it
    outputs the first variant of a sum type.
    """

    #: Types that are only inputs of the child graph.
    just_inputs: tys.TypeRow
    #: Types that are appended to both inputs and outputs of the graph.
    rest: tys.TypeRow
    _just_outputs: tys.TypeRow | None = field(default=None, repr=False)
    extension_delta: tys.ExtensionSet = field(default_factory=list, repr=False)

    @property
    def just_outputs(self) -> tys.TypeRow:
        """Types that are only outputs of the child graph.

        Raises:
            IncompleteOp: If the outputs have not been set.
        """
        return _check_complete(self, self._just_outputs)

    @property
    def num_out(self) -> int:
        return len(self.just_outputs) + len(self.rest)

    def to_serial(self, parent: Node) -> sops.TailLoop:
        return sops.TailLoop(
            parent=parent.idx,
            just_inputs=ser_it(self.just_inputs),
            just_outputs=ser_it(self.just_outputs),
            rest=ser_it(self.rest),
            extension_delta=self.extension_delta,
        )

    def inner_signature(self) -> tys.FunctionType:
        return tys.FunctionType(
            self._inputs(), [tys.Sum([self.just_inputs, self.just_outputs]), *self.rest]
        )

    def outer_signature(self) -> tys.FunctionType:
        return tys.FunctionType(self._inputs(), self.just_outputs + self.rest)

    def _set_out_types(self, types: tys.TypeRow) -> None:
        (sum_, other) = tys.get_first_sum(types)
        just_ins, just_outs = sum_.variant_rows
        assert (
            just_ins == self.just_inputs
        ), "First sum variant rows don't match TailLoop inputs."
        self._just_outputs = just_outs

    def _inputs(self) -> tys.TypeRow:
        return self.just_inputs + self.rest


@dataclass
class FuncDefn(DfParentOp):
    """Function definition operation, parent of a dataflow graph that defines
    the function.
    """

    #: function name
    name: str
    #: input types of the function
    inputs: tys.TypeRow
    # ? type parameters of the function if polymorphic
    params: list[tys.TypeParam] = field(default_factory=list)
    _outputs: tys.TypeRow | None = field(default=None, repr=False)
    num_out: int = field(default=1, repr=False)

    @property
    def outputs(self) -> tys.TypeRow:
        """Output types of the function.

        Raises:
            IncompleteOp: If the outputs have not been set.
        """
        return _check_complete(self, self._outputs)

    @property
    def signature(self) -> tys.PolyFuncType:
        """Polymorphic signature of the function.

        Raises:
            IncompleteOp: If the outputs have not been set.
        """
        return tys.PolyFuncType(
            self.params, tys.FunctionType(self.inputs, self.outputs)
        )

    def to_serial(self, parent: Node) -> sops.FuncDefn:
        return sops.FuncDefn(
            parent=parent.idx,
            name=self.name,
            signature=self.signature.to_serial(),
        )

    def inner_signature(self) -> tys.FunctionType:
        return self.signature.body

    def _set_out_types(self, types: tys.TypeRow) -> None:
        self._outputs = types

    def _inputs(self) -> tys.TypeRow:
        return self.inputs

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        match port:
            case OutPort(_, 0):
                return tys.FunctionKind(self.signature)
            case _:
                raise self._invalid_port(port)


@dataclass
class FuncDecl(Op):
    """Function declaration operation, defines the signature of a function."""

    #: function name
    name: str
    #: polymorphic function signature
    signature: tys.PolyFuncType
    num_out: int = field(default=1, repr=False)

    def to_serial(self, parent: Node) -> sops.FuncDecl:
        return sops.FuncDecl(
            parent=parent.idx,
            name=self.name,
            signature=self.signature.to_serial(),
        )

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        match port:
            case OutPort(_, 0):
                return tys.FunctionKind(self.signature)
            case _:
                raise self._invalid_port(port)


@dataclass
class Module(Op):
    """Root operation of a HUGR which corresponds to a full module definition."""

    num_out: int = field(default=0, repr=False)

    def to_serial(self, parent: Node) -> sops.Module:
        return sops.Module(parent=parent.idx)

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        raise self._invalid_port(port)


class NoConcreteFunc(Exception):
    """Could not instantiate a polymorphic function."""


@dataclass
class _CallOrLoad:
    #: polymorphic function signature
    signature: tys.PolyFuncType
    #: concrete function signature
    instantiation: tys.FunctionType
    #: type arguments for polymorphic function
    type_args: list[tys.TypeArg]

    def __init__(
        self,
        signature: tys.PolyFuncType,
        instantiation: tys.FunctionType | None = None,
        type_args: Sequence[tys.TypeArg] | None = None,
    ) -> None:
        self.signature = signature

        if len(signature.params) == 0:
            self.instantiation = signature.body
            self.type_args = []

        else:
            # TODO substitute type args into signature to get instantiation
            if instantiation is None:
                msg = "Missing instantiation for polymorphic function."
                raise NoConcreteFunc(msg)
            type_args = type_args or []

            if len(signature.params) != len(type_args):
                msg = "Mismatched number of type arguments."
                raise NoConcreteFunc(msg)
            self.instantiation = instantiation
            self.type_args = list(type_args)


class Call(_CallOrLoad, Op):
    """Call a function inside a dataflow graph. Connects to :class:`FuncDefn` or
    :class:`FuncDecl` operations.

    Args:
        signature: Polymorphic function signature.
        instantiation: Concrete function signature. Defaults to None.
        type_args: Type arguments for polymorphic function. Defaults to None.

    Raises:
        NoConcreteFunc: If the signature is polymorphic and no instantiation
            is provided.
    """

    def to_serial(self, parent: Node) -> sops.Call:
        return sops.Call(
            parent=parent.idx,
            func_sig=self.signature.to_serial(),
            type_args=ser_it(self.type_args),
            instantiation=self.instantiation.to_serial(),
        )

    @property
    def num_out(self) -> int:
        return len(self.signature.body.output)

    def _function_port_offset(self) -> int:
        return len(self.signature.body.input)

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        match port:
            case InPort(_, offset) if offset == self._function_port_offset():
                return tys.FunctionKind(self.signature)
            case _:
                return tys.ValueKind(_sig_port_type(self.instantiation, port))


@dataclass()
class CallIndirect(DataflowOp, _PartialOp):
    """Higher order evaluation of a
    :class:`FunctionType <hugr.tys.FunctionType>` value.
    """

    _signature: tys.FunctionType | None = None

    @property
    def num_out(self) -> int:
        return len(self.signature.output)

    @property
    def signature(self) -> tys.FunctionType:
        """The signature of the function being called.

        Raises:
        IncompleteOp: If the signature has not been set.
        """
        return _check_complete(self, self._signature)

    def to_serial(self, parent: Node) -> sops.CallIndirect:
        return sops.CallIndirect(
            parent=parent.idx,
            signature=self.signature.to_serial(),
        )

    def __call__(self, function: Wire, *args: Wire) -> Command:  # type: ignore[override]
        return super().__call__(function, *args)

    def outer_signature(self) -> tys.FunctionType:
        sig = self.signature

        return tys.FunctionType(input=[sig, *sig.input], output=sig.output)

    def _set_in_types(self, types: tys.TypeRow) -> None:
        func_sig, *_ = types
        assert isinstance(
            func_sig, tys.FunctionType
        ), f"Expected function type, got {func_sig}"
        self._signature = func_sig


class LoadFunc(_CallOrLoad, DataflowOp):
    """Load a statically defined function as a higher order value.
      Connects to :class:`FuncDefn` or :class:`FuncDecl` operations.

    Args:
        signature: Polymorphic function signature.
        instantiation: Concrete function signature. Defaults to None.
        type_args: Type arguments for polymorphic function. Defaults to None.

    Raises:
        NoConcreteFunc: If the signature is polymorphic and no instantiation
            is provided.
    """

    num_out: int = field(default=1, repr=False)

    def to_serial(self, parent: Node) -> sops.LoadFunction:
        return sops.LoadFunction(
            parent=parent.idx,
            func_sig=self.signature.to_serial(),
            type_args=ser_it(self.type_args),
            signature=self.outer_signature().to_serial(),
        )

    def outer_signature(self) -> tys.FunctionType:
        return tys.FunctionType(input=[], output=[self.instantiation])

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        match port:
            case InPort(_, 0):
                return tys.FunctionKind(self.signature)
            case OutPort(_, 0):
                return tys.ValueKind(self.instantiation)
            case _:
                raise self._invalid_port(port)


@dataclass
class Noop(DataflowOp, _PartialOp):
    """Identity operation that passes through its input."""

    _type: tys.Type | None = None
    num_out: int = field(default=1, repr=False)

    @property
    def type_(self) -> tys.Type:
        """The type of the input and output of the operation."""
        return _check_complete(self, self._type)

    def to_serial(self, parent: Node) -> sops.Noop:
        return sops.Noop(parent=parent.idx, ty=self.type_.to_serial_root())

    def outer_signature(self) -> tys.FunctionType:
        return tys.FunctionType.endo([self.type_])

    def _set_in_types(self, types: tys.TypeRow) -> None:
        (t,) = types
        self._type = t

    def __repr__(self) -> str:
        return "Noop" + (f"({self._type})" if self._type is not None else "")


@dataclass
class Lift(DataflowOp, _PartialOp):
    """Add an extension requirement to input values and pass them through."""

    #: Extension added.
    new_extension: tys.ExtensionId
    _type_row: tys.TypeRow | None = field(default=None, repr=False)
    num_out: int = field(default=1, repr=False)

    @property
    def type_row(self) -> tys.TypeRow:
        """Types of the input and output of the operation."""
        return _check_complete(self, self._type_row)

    def to_serial(self, parent: Node) -> sops.Lift:
        return sops.Lift(
            parent=parent.idx,
            new_extension=self.new_extension,
            type_row=ser_it(self.type_row),
        )

    def outer_signature(self) -> tys.FunctionType:
        return tys.FunctionType.endo(self.type_row)

    def _set_in_types(self, types: tys.TypeRow) -> None:
        self._type_row = types


@dataclass
class AliasDecl(Op):
    """Declare an external type alias."""

    #: Alias name.
    name: str
    #: Type bound.
    bound: tys.TypeBound
    num_out: int = field(default=0, repr=False)

    def to_serial(self, parent: Node) -> sops.AliasDecl:
        return sops.AliasDecl(
            parent=parent.idx,
            name=self.name,
            bound=self.bound,
        )

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        raise self._invalid_port(port)


@dataclass
class AliasDefn(Op):
    """Declare a type alias."""

    #: Alias name.
    name: str
    #: Type definition.
    definition: tys.Type
    num_out: int = field(default=0, repr=False)

    def to_serial(self, parent: Node) -> sops.AliasDefn:
        return sops.AliasDefn(
            parent=parent.idx,
            name=self.name,
            definition=self.definition.to_serial_root(),
        )

    def port_kind(self, port: InPort | OutPort) -> tys.Kind:
        raise self._invalid_port(port)



"""Builder classes for defining functions and modules in HUGR."""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from hugr import ops, val

from .dfg import _DfBase
from .hugr import Hugr

if TYPE_CHECKING:
    from hugr.node_port import Node

    from .tys import PolyFuncType, Type, TypeBound, TypeParam, TypeRow


@dataclass
class Function(_DfBase[ops.FuncDefn]):
    """Build a function definition as a HUGR dataflow graph.

    Args:
        name: The name of the function.
        input_types: The input types for the function (output types are
        computed by propagating types from input node through the graph).
        type_params: The type parameters for the function, if polymorphic.

    Examples:
        >>> f = Function("f", [tys.Bool])
        >>> f.parent_op
        FuncDefn(name='f', inputs=[Bool], params=[])
    """

    def __init__(
        self,
        name: str,
        input_types: TypeRow,
        type_params: list[TypeParam] | None = None,
    ) -> None:
        root_op = ops.FuncDefn(name, input_types, type_params or [])
        super().__init__(root_op)


@dataclass
class Module:
    """Build a top-level HUGR module.

    Examples:
        >>> m = Module()
        >>> m.hugr.root_op()
        Module()
    """

    hugr: Hugr[ops.Module]

    def __init__(self) -> None:
        self.hugr = Hugr(ops.Module())

    def define_function(
        self,
        name: str,
        input_types: TypeRow,
        type_params: list[TypeParam] | None = None,
    ) -> Function:
        """Start building a function definition in the module.

        Args:
            name: The name of the function.
            input_types: The input types for the function.
            type_params: The type parameters for the function, if polymorphic.

        Returns:
            The new function builder.
        """
        parent_op = ops.FuncDefn(name, input_types, type_params or [])
        return Function.new_nested(parent_op, self.hugr)

    def define_main(self, input_types: TypeRow) -> Function:
        """Define the 'main' function in the module. See :meth:`define_function`."""
        return self.define_function("main", input_types)

    def declare_function(self, name: str, signature: PolyFuncType) -> Node:
        """Add a function declaration to the module.

        Args:
            name: The name of the function.
            signature: The (polymorphic) signature of the function.

        Returns:
            The node representing the function declaration.

        Examples:
            >>> m = Module()
            >>> sig = tys.PolyFuncType([], tys.FunctionType.empty())
            >>> m.declare_function("f", sig)
            Node(1)
        """
        return self.hugr.add_node(ops.FuncDecl(name, signature), self.hugr.root)

    def add_const(self, value: val.Value) -> Node:
        """Add a static constant to the module.

        Args:
            value: The constant value to add.

        Returns:
            The node holding the constant.

        Examples:
            >>> m = Module()
            >>> m.add_const(val.FALSE)
            Node(1)
        """
        return self.hugr.add_node(ops.Const(value), self.hugr.root)

    def add_alias_defn(self, name: str, ty: Type) -> Node:
        """Add a type alias definition."""
        return self.hugr.add_node(ops.AliasDefn(name, ty), self.hugr.root)

    def add_alias_decl(self, name: str, bound: TypeBound) -> Node:
        """Add a type alias declaration."""
        return self.hugr.add_node(ops.AliasDecl(name, bound), self.hugr.root)
